% ANNEAL_EX  An example of simulated annealing for f(x) functions.
% It contains implementation of Boltzmann and Cauchy annelaing and
% two test functions: Rastrigin function and polynom from [1]
%
% [1] Szu H., Hartley R. Fast simulated annealing //
% // Phys. Lett. A. 1987. V.122. N3-4. P.157-162.
function anneal_ex
% Select method and initial conditions
funcid = 'rast'; % rast --- Rastrigin function, polypol -- polynomial
method = 2; % 1 -- Boltzmann, 2 --- Cauchy
if strcmp(funcid, 'rast')
    if method == 1
        T0 = 10;
        iternum = 200000;
    else
        T0 = 1000;
        iternum = 200000;
    end
    func = @rast;
    x = 20000;
elseif strcmp(funcid, 'polypot')
    if method == 1
        T0 = 10;
        iternum = 250;
    else
        T0 = 10;
        iternum = 250;
    end
    func = @polypot;
    x = 3;
end
% Iniialize arrays for diagrams
x_v = nan(1,iternum);
xglob_v = nan(1,iternum);
xglob = Inf; fglob = Inf;
% Simulated annealing method iterations
for k = 1:iternum
    if method == 1 % a) Boltzmann annealing
        T = T0 / log(k + 1);
        xnew = x + T*randn(1,1);
    else % b) Cauchy annealing
        T = T0 / k;
        xnew = x + T*tan(pi*(rand(1,1) - 0.5));
    end
    % Calculate function values
    fnew = func(xnew);
    f = func(x);
    % Check if new function value is the best solution
    if fnew < fglob;        
        xglob = xnew;
        fglob = fnew;
    end
    % Add data to arrays for diagrams construction
    x_v(k) = x;
    xglob_v(k) = xglob;
    % Solve if the old solution will be replaced to the new solution
    if fnew < f
        x = xnew;
    else
        p = exp(-(fnew - f)./T);
        if rand(1,1) < p
            x = xnew;
        end
    end
    % Debug print
    if mod(k,iternum / 100) == 0
        fprintf('Iteration %d; x = %f, T = %f, xglob = %f\n', k, x, T, xglob);
    end
end
% Reduce size of points array by calculating mean values
n = 25;
xglob_sv = mean(reshape(abs(xglob_v), n, iternum/n), 1);
x_sv = mean(reshape(abs(x_v), n, iternum/n), 1);
% Draw graphs
semilogy(n*(1:numel(x_sv))/1000, x_sv, 'b-', 'LineWidth', 2)
hold on;
semilogy(n*(1:numel(xglob_sv))/1000, xglob_sv, 'r-', 'LineWidth', 2);
hold off;
set(gca, 'FontSize', 12);
xlabel('n/1000', 'FontSize', 14);
ylabel('F(x)', 'FontSize', 14);
end

function y = rast(x)
y = x.^2 + 10 - 10*cos(2*pi*x);
end

function y = polypot(x)
y = x.^4 - 16*x.^2 + 5*x;
end