% GENETIC_MINSEARCH  Finds the minimum of the function using genetic
% algorithm. Some options are included directly into code for simplicity
%
% Usage:
% [x, fval] = genetic_minsearch(func, x0)
% Inputs:
%   func -- @(x) (...) anonymous function handle
%   x0 -- initial approximation (row vector)
% Outputs
%   x -- value of x at the minimum
%   fval -- f(x) function value at the minimum
function [x, fval] = genetic_minsearch(func, x0)
% Some internal settings
sigma_far = 1; % Variance value for the 
sigma_near = 0.1;
pop_size = 1000; % Population size
elite_size = 50; % Elite size
max_iter = 100; % Maximum number of iteration
visualize = true; % Visualize each iteration on 2D graph (for demo)
% Create initial population
x0 = x0(:)';
xmat = repmat(x0, pop_size, 1) + sigma_far*randn(pop_size, numel(x0));
fvals = nan(pop_size, 1);
for i = 1:pop_size
    fvals(i) = func(xmat(i,:));
end
% Find the fittest x vector
for iter=1:max_iter
    % Select the fittest x vectors ("elite")
    [fvals, ind] = sort(fvals);
    xmat = xmat(ind,:);
    fprintf('Iteration %d: fval=%e; b = [', iter, fvals(1));    
    fprintf('%e ', xmat(1,:)); fprintf(']\n');    
    % Visualization (mainly for lecture)
    if visualize
        plot(xmat((elite_size+1):pop_size,1), xmat((elite_size+1):pop_size,2), 'b.', 'LineWidth', 2);        
        hold on;
        plot(xmat(1:elite_size,1), xmat(1:elite_size,2), 'ro', 'LineWidth', 2);        
        hold off;
        axis([-10 10 -10 10]);    
        grid on;
        box on;
        pause(0.25);
        drawnow;
    end    
    % Create new generation
    for i=(elite_size+1):pop_size
        % Select "parents"
        pind1 = mod(round(1e6*rand(1,1)), elite_size) + 1;
        pind2 = mod(round(1e6*rand(1,1)), elite_size) + 1;
        % Generate "weight" vector
        w = rand(1, numel(x0));
        % Create "child" vector
        % a) crossingover
        xmat(i,:) = (1-w).*xmat(pind1,:) + w.*xmat(pind2,:);
        % b) mutation
        mut_prob = rand(1,1);        
        if mut_prob < 0.1 % Near-range mutation
            xmat(i,:) = xmat(i,:) + sigma_near.*randn(1, numel(x0));
        elseif mut_prob < 0.2 % Far-range mutation
            xmat(i,:) = xmat(i,:) + sigma_far.*randn(1, numel(x0));
        elseif mut_prob < 0.3 % Mutation that uses variance inside elite
            sigma_cust = std(xmat(:,1),[],1);            
            xmat(i,:) = xmat(i,:) + sigma_cust.*randn(1, numel(x0));
        end
        fvals(i) = func(xmat(i,:));                    
    end
end
% Return the result
x = xmat(1,:);
fval = fvals(2,:);
end