% Genetic algorithm global optimization demo (on the example of
% Rastrigin function). It includes demonstration of three methods applied
% to 2D (f(x,y)) Rastrigin function:
% 1) Levenberg-Marquardt (as an example of optimizator failure)
% 2) Simulated annealing
% 3) Genetic algorithm (with interactive animation)
%
% See the next reference for another test functions
% http://www.sfu.ca/~ssurjano/egg.html
function globoptim_ex
close all;
plots_to_file = false;
% Show Rastrigin function graph
% a) contour plot
[X,Y] = meshgrid(-3:0.1:3, -3:0.1:3);
Z = rastrigin(X,Y);
contour(X,Y,Z, 'LineWidth', 2);
xlabel('{\it{x}}_1','FontSize',18);
ylabel('{\it{x}}_2','FontSize',18);
set(gca, 'FontSize', 16);
box on;
if plots_to_file
    print('ch_glopt_rastc', '-dpng', '-r600');
end
% b) surface plot
figure;
[X,Y] = meshgrid(-5:0.02:5, -5:0.02:5);
Z = rastrigin(X,Y);
xlabel('{\it{x}}_1','FontSize',18);
ylabel('{\it{x}}_2','FontSize',18);
zlabel('{\it{f}}({\it{x}},{\it{y}})', 'FontSize', 18);
surf(X,Y,Z, 'LineStyle', 'none');
colormap(jet);
set(gca, 'FontSize', 16);
box on;
if plots_to_file
    print('ch_glopt_rasts', '-dpng', '-r600');
end
fprintf('Press <ENTER> to launch Levenberg-Marquardt optimization\n');
pause;
% Run Levenberg-Marquardt optimization procedure
b0 = [10 10];
b = lsqfit_lm(0, 0, @(b,x) rastrigin(b(1),b(2)), b0)
fprintf('Press <ENTER> to launch global optimization algorithm\n');
pause;
% Run global optimization method
method = menu('Select global optimization method','Simulated annealing', 'Genetic algorithm');
if method == 1
    % Run simulated annealing
    anneal;
else
    % Run genetic algorithm optimization procedure
    figure;
    b0 = [10 10];
    [b, fval] = genetic_minsearch(@(b) rastrigin(b(1),b(2)), b0)
end
end

function anneal
% Simulated annealing for Rastrigin function
b0 = [10; 10];
T0 = 10;
fbest = Inf;
bbest = [NaN NaN];
for n=1:5000
    % a) Cauchy annealing
    T = T0 / n;
    b = b0 + T*trnd(1);
    % b) Boltzmann annealing
    %T = T0 / log(1 + n);
    %b = b0 + T*randn(1,1);
    
    f0 = rastrigin(b0(1),b0(2));
    if f0 < fbest
        bbest = b0;
        fbest = f0;
    end    
    f_new = rastrigin(b(1),b(2));
    p = exp(-(f_new - f0)./T);
    if mod(n,125) == 0
        fprintf('%5d) b=[%8.5f %8.5f], f=%5.3f, T=%5.3f; ', n, b, f_new, T);
        fprintf('b_best=[%8.5f %8.5f]\n', bbest);
    end
    pctl = rand(1,1);
    if pctl < p
        b0 = b;
    end
end
end

function [F, J] = rastrigin(x1,x2)
% Rastrigin function (2D case) implementation
F = 20 + x1.^2 - 10*cos(2*pi*x1) + x2.^2 - 10*cos(2*pi*x2);
if nargout == 2
    J = [2*x1 + 20*pi*sin(2*pi*x1), 2*x2 + 20*pi*sin(2*pi*x2)];
end
end