% LSQFIT_LM  Simple implementation of Levenberg-Marquardt method for
% non-linear regression.
%
% Usage:
%   [b,res,J] = lsqfit_lm(X,Y,func,b0)
% Inputs:
%   X -- independent variable(s) matrix for function func
%   Y -- dependent variable vector for function func
%   func --- anonymous function: [F,J] = func(b,X)
%            F and J are Nx1 and NxM func.values and Jacobian respectively
%            b is Mx1 function parameters vector
% Outputs:
%   b -- optimized coefficients
%   res -- Nx1 regression residuals vector
%   J -- NxM Jacobian of the function for b and X values
%
% See also LSQFIT_LM, LSQFIT_CI
function [b, res, J] = lsqfit_lm(X, Y, func, b0)
% Check input parameters sizes
Y = Y(:); % Convert Y to column
if size(X, 1) == 1
    X = X';
end
if size(X, 1) ~= numel(Y)
    error('X and Y sizes are not consistent');
end
% Some initial parameters
disp('Levenberg-Marquardt method based non-linear regression solver');
delta = 1e-15; % Solver accuracy (b tolerance)
ftol = 1e-15; % Solver accuracy (function tolerance)
lmin = 1e-15; % Minimal regularization parameter value
b0 = b0(:); % Transform to column vector
m = numel(b0);
db = ones(size(b0)); % Initial value (to run initial iteration)
dF = 1;
while sqrt((db'*db)/m) > delta & dF > ftol
    % Calculate Jacobian and initial function value
    [F, J] = func(b0, X);
    F = F - Y;
    % LM algorithm step finding
    l = lmin; % Initial step
    Fnew = 1.1 * F;
    while Fnew'*Fnew > F'*F
        A = J'*J + l*eye(m);
        if rcond(A) > eps
            db = A \ (-J'*F);
            if any(isnan(db) | isinf(db))
                error('Ill-conditioned equation system in the iteration');
            end        
            b0_new = b0 + db;
            Fnew = func(b0_new, X) - Y;
        end
        l = l * 4; % Increase step        
    end
    dF = abs(F - Fnew);
    % Accept new approximation and show the result
    b0 = b0_new; F = Fnew;
    fprintf('b=['); fprintf('%.5e ', b0); fprintf('] ');
    fprintf('F=%.5e l=%.3e\n', F'*F, l);
end
b = b0; res = F;
end